-- /*******************************************************************************

--   Copyright (c) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata
--   All rights reserved.

-- ********************************************************************************

--   Redistribution and use in source and binary forms, with or without 
--   modification, are permitted provided that the following conditions are met:
--     * Redistributions of source code must retain the above copyright
--       notice, this list of conditions and the following disclaimer.
--     * Redistributions in binary form must reproduce the above copyright
--       notice, this list of conditions and the following disclaimer in the
--       documentation and/or other materials provided with the distribution.
--     * Neither the name of the author nor the names of its contributors may
--       be used to endorse or promote products derived from this software
--       without specific prior written permission.

--   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
--   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
--   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
--   DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
--   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
--   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
--   OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
--   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
--   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
--   EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

-- ********************************************************************************

--   La redistribución y el uso en las formas de código fuente y binario, con o sin
--   modificaciones, están permitidos siempre que se cumplan las siguientes condiciones:
--     * Las redistribuciones del código fuente deben conservar el aviso de copyright
--       anterior, esta lista de condiciones y el siguiente descargo de responsabilidad.
--     * Las redistribuciones en formato binario deben reproducir el aviso de copyright
--       anterior, esta lista de condiciones y el siguiente descargo de responsabilidad
--       en la documentación y/u otros materiales suministrados con la distribución.
--     * Ni el nombre de los titulares de derechos de autor ni los nombres de sus
--       colaboradores pueden usarse para apoyar o promocionar productos derivados de
--       este software sin permiso previo y por escrito.

--   ESTE SOFTWARE SE SUMINISTRA POR LIFIA "COMO ESTÁ" Y CUALQUIER GARANTÍA EXPRESA
--   O IMPLÍCITA, INCLUYENDO, PERO NO LIMITADO A, LAS GARANTÍAS IMPLÍCITAS DE
--   COMERCIALIZACIÓN Y APTITUD PARA UN PROPÓSITO DETERMINADO SON RECHAZADAS. EN
--   NINGÚN CASO LIFIA SERÁ RESPONSABLE POR NINGÚN DAÑO DIRECTO, INDIRECTO, INCIDENTAL,
--   ESPECIAL, EJEMPLAR O CONSECUENTE (INCLUYENDO, PERO NO LIMITADO A, LA ADQUISICIÓN
--   DE BIENES O SERVICIOS; LA PÉRDIDA DE USO, DE DATOS O DE BENEFICIOS; O INTERRUPCIÓN
--   DE LA ACTIVIDAD EMPRESARIAL) O POR CUALQUIER TEORÍA DE RESPONSABILIDAD, YA SEA
--   POR CONTRATO, RESPONSABILIDAD ESTRICTA O AGRAVIO (INCLUYENDO NEGLIGENCIA O
--   CUALQUIER OTRA CAUSA) QUE SURJA DE CUALQUIER MANERA DEL USO DE ESTE SOFTWARE,
--   INCLUSO SI SE HA ADVERTIDO DE LA POSIBILIDAD DE TALES DAÑOS.

-- *******************************************************************************/

local log = require 'log'

----------------------------------------------------------------------------------

-- Internal vars
local M = {}
local status = { enabled = 0, disabled = 1, running = 2, setuping = 3}

----------------------------------------------------------------------------------
-- Application API
----------------------------------------------------------------------------------

function M.new( id, environment, manager, surface )
	local app = {}
	app.id = id
	app.env = environment
	app._manager = manager
	app._surface = surface

	app.status = status.disabled

	return app
end

function M.initialize( app )
	local desc = callAppMethod(app.env, app.env.describe, {})
	if check_description(desc) then
		app.methods = desc.methods
		app.name = desc.name
		app.version = desc.version
		app.desc = desc.desc
		app.interactive = desc.interactive

		log.debug("Application", "initialized: name= "  .. app.name)

		return true
	end
	return false
end

function M.finalize( app )
end

function M.enable( app )
	local result = true

	if app.interactive then
		app.status = status.enabled
	else
		result = M.start(app)
	end

	return result
end

function M.disable( app )
	if app.interactive then
		app.status = status.disabled
	else
		M.stop(app)
	end
end

function M.start( app )
	local result = false

	if app.interactive then
		result = start_interactive(app)
	else
		result = start_service(app)
	end

	if result then
		app.status = status.running
	end

	return result
end

function M.stop( app )
	dispatchMethod(app, app.methods.stop)
	if app.interactive then
		stop_interactive(app)
	else
		stop_service(app)
	end
end

function M.start_setup( app )
	if dispatchMethod(app, app.methods.startSetup, {app._surface}) then
		start_window(app._manager, app.id)
		app.status = status.setuping
		zapperApplicationLayerEnabled( 1 )
	end
end

function M.stop_setup( app )
	dispatchMethod(app, app.methods.stopSetup)
	stop_window(app._manager)

	if app.interactive then
		app.status = status.enabled
	else
		app.status = status.running
	end
	zapperApplicationLayerEnabled( 0 )
end

function M.close( app )
	if app.status == status.setuping then
		M.stop_setup(app)
	else
		M.stop(app)
	end
end

function M.is_running( app )
	return app.status == status.running
end

function M.is_enable( app )
	return app.status ~= status.disabled
end

function M.is_interactive( app )
	return app.interactive
end

function M.is_service( app )
	return not M.is_interactive(app)
end

function M.is_setupable( app )
	return app.methods.startSetup ~= nil
end


function M.handle_event( app, evt )
	dispatchMethod(app, app.methods.handleEvent, {evt})
end

----------------------------------------------------------------------------------
-- Application Internals
----------------------------------------------------------------------------------

-- dispatchMethod(app, method,[ [arguments], [abort] ])
-- arguments is a table with values that are going to pass to the method as arguments
-- abort specify if the app must be stopped when the method fails or return false
function dispatchMethod( app, method, args, abort )
	local result = false

	-- If args is missing ( args must be a table ) then abort takes the value
	if type(args) == "boolean" then
		abort = args
	end

	if method ~= nil then
		if callAppMethod(app.env, method, args) then
			result = true
		elseif abort then
			M.stop(app)
		end
	else
		local msg = "ZaMBA application does not implement the method"
		if abort then
			log.warn( "Application", msg )
		else
			log.info( "Application", msg )
		end
	end

	return result
end

function callAppMethod( env, method, args )
	--Fix args
	if type(args) ~= "table" then args = {args} end

	local methodResult = nil
	local result, errorMsg = pcall( function() methodResult = method(unpack(args)) end)

	if (result) then
		return methodResult
	else
		log.warn("Application", "ZaMBA application generated an error: " .. errorMsg)
		return false
	end
end

function check_description( desc )
	local result = type(desc) == "table"

	if result then
		result = (result and (desc.name ~= nil))
		result = (result and (type(desc.interactive) == "boolean"))
		result = (result and (type(desc.methods) == "table"))

		if result then
			local methods = desc.methods
			result = (result and (type(methods.start) == "function"))
			result = (result and (type(methods.stop) == "function"))
			result = (result and (type(methods.handleEvent) == "function"))
		end
	end

	return result
end

-- Interactive methods

function start_interactive( app )
	if not is_running_interactive(app._manager) then
		if dispatchMethod(app, app.methods.start, {app._surface}, true) then
			start_window(app._manager, app.id)
			zapperApplicationLayerEnabled( 1 )
			return true
		end
	end
	return false
end

function stop_interactive( app )
	stop_window(app._manager)
	app.status = status.enabled
	zapperApplicationLayerEnabled( 0 )
end

-- Service methods

function start_service( app )
	return dispatchMethod(app, app.methods.start, true)
end

function stop_service( app )
	app.status = status.disabled
end

return M