local l_util = require 'l_util'
local log = require 'log'

_G.name = "GLOBAL"

function makePath( basepath, filepath )
	local t = l_util.string.split(filepath, ".")
	local tmp = ""
	for _, p in ipairs(t) do
		tmp = util.makePath(tmp, p)
	end
	if tmp:find("%.lua$") == nil then
		tmp = tmp .. ".lua"
	end
	return util.makePath(basepath, tmp)
end

function require_in_environment( modname, env )
	local name = makePath(env.rootPath, modname)

	-- Lookup in the cache
	if env.package.loaded[name] ~= nil then
		return env.package.loaded[name]
	end

	local result, err = load_chunk(name, env)

	-- Store result in the cache
	if err == nil then
		log.debug("Sandbox", "Require local module " .. modname .. " sucessful")
		env.package.loaded[name] = result
	else
		if not_found_error(err) then
			log.debug("Sandbox", "Module " .. modname .. " not found in app context... will be performed a lookup in the system")
			result = require(modname)
		else
			log.error("Sandbox", err)
		end
	end

	return result
end

function not_found_error( error_str )
	return string.find(error_str, "cannot open") ~= nil
end

function create_env( configure, ... )
	local env = {}
	setmetatable(env, {__index = _G})

	env.name = "SANDBOX"
	env._G = env

	env.module = function() return nil end

	env.package = {
		loaded = {},
		loaders = {},
		preload = {},
		path = nil,
		cpath = nil,
		seeall = function(...) return nil end,
		loadlib = function(...) return nil end
	}

	-- Require is made in the SANDBOX environment
	env.require = function( modname ) return require_in_environment(modname, env) end

	if configure ~= nil then
		setfenv(configure, env)(_G, ...)
	end

	return env
end

function load_chunk(file, env)
	local result = false
	local chunk, err = loadfile(file)

	if err == nil then
		result = setfenv(chunk, env)()
	end
	return result, err
end

function sandbox( file, configure, ... )
	local env = create_env( configure, ... )

	local result, err = load_chunk(file, env)
	if err ~= nil then
		return err
	end
	return env
end

function wrap( fnc, new_env )
	return function(...)
		old_env = _G
		setfenv(0, new_env)
		local r = fnc(...)
		setfenv(0, old_env)
		return r
	end
end

return { sandbox = sandbox, wrap = wrap }

